#!/usr/bin/bash

# Define a timestamp function
timestamp() {
  date +"%Y-%m-%dT%T.%3N%z" # current time
}

echo "Copying project $1"

#current working directory
ROOT_DIR=/DyPyBench

cd "$3" || exit

#copy project folder to temp folder
cp -r "$ROOT_DIR/../Project/project$2" .

cd "project$2" || exit
