import logging

from synthegrator.randomstuff.getk import get_openai_key


def test_oai():
    from langchain.llms import OpenAI

    llm = OpenAI(openai_api_key=get_openai_key())
    logging.info(llm)
    text = "What would be a good company name for a company that makes colorful socks?"
    result = llm.predict(text)
    logging.info(result)
    logging.info(type(result))


def main():
    test_oai()


if __name__ == "__main__":
    main()
