def main():
    from pprint import pprint

    from datasets import load_dataset

    dataset = load_dataset("THUDM/humaneval-x", "java")
    for ex in dataset["test"]:
        pprint(ex)
        print("Prompt ---")
        print(ex["prompt"])
        print("Canonical_solution ---")
        print(ex["canonical_solution"])
        print("Test cases ---")
        print(ex["test"])
        break


if __name__ == "__main__":
    main()
